"""Public key registration app."""

from enum import Enum
from typing import Annotated, Optional

import typer
import yaml
from pydantic import RootModel
from rich.table import Table
from slices_pubkey_client.exceptions import ApiException, NotFoundException
from slices_pubkey_client.models.public_key import PublicKey
from slices_pubkey_client.models.public_key_create import PublicKeyCreate

from . import console, err_console
from .context import Context
from .users_info import USER_ID_HELP
from .utils import get_error_detail

pubkey_app = typer.Typer(name="pubkey", help="SSH key management", no_args_is_help=True)


@pubkey_app.command(name="register", rich_help_panel="Commands")
def register_pubkey(ctx: Context, file: Annotated[typer.FileText, typer.Argument]):
    """Register public key."""
    pubkey_api = ctx.obj.api.pubkey_api

    pubkey = file.read()

    PublicKeyCreate(public_key=pubkey)

    try:
        pubkey_api.register_user_pubkey(
            ctx.obj.get_logged_in_userinfo("sub"),
            PublicKeyCreate(public_key=pubkey),
        )
        console.print("[bold green]Registered public key.[/] :white_check_mark:")
    except ApiException as err:
        err_console.print("[bold red]Could not register public key[/] :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(code=1) from None


class Format(Enum):
    """Output format Options."""

    table = "table"
    text = "text"
    json = "json"
    yaml = "yaml"


PublicKeys = RootModel[list[PublicKey]]


def validate_user_id(ctx: typer.Context, value: str):
    """Validate user ID, resolving it if it is not a TypeID. Always return TypeID."""
    if value is None:
        return None
    return ctx.obj.other_users_info.resolve_project_member(ctx.obj.require_project_id(), value)


@pubkey_app.command(name="list", rich_help_panel="Commands")
def list_pubkeys(
    ctx: Context,
    user_id: Annotated[
        Optional[str], typer.Option(help=f"User ID to list public keys for. {USER_ID_HELP}", callback=validate_user_id)
    ] = None,
    fingerprint: Annotated[bool, typer.Option(help="Show pubkey fingerprint instead of full key.")] = True,
    format: Annotated[Format, typer.Option("--format", "-f", help="Output format.")] = "table",  # type: ignore
):
    """List registered public key."""
    pubkey_api = ctx.obj.api.pubkey_api

    is_local_user = not user_id
    if not user_id:
        user_id = ctx.obj.get_logged_in_userinfo("sub")

    pubkeys = pubkey_api.get_user_pubkeys(user_id)

    if format == Format.table:
        table = Table("ID", "Fingerprint" if fingerprint else "Public Key")
        for pubkey in pubkeys.public_keys:
            table.add_row(str(pubkey.id), pubkey.sha256_fingerprint if fingerprint else pubkey.public_key)

        if not is_local_user:
            user_info = ctx.obj.other_users_info.try_get_user_info_opt(user_id)
            console.print(f"Public keys for {user_id}")
            if user_info:
                console.print(
                    f"User details: {user_info.first_name} {user_info.last_name} - "
                    f"{user_info.email} - {user_info.preferred_username}"
                )
        console.print(table)
        console.print(f"{pubkeys.total} public keys registered")
    elif format == Format.text:
        for pubkey in pubkeys.public_keys:
            console.print(pubkey.public_key)
    elif format == Format.json:
        console.print_json(PublicKeys(pubkeys.public_keys).model_dump_json(by_alias=True, exclude_unset=True))
    elif format == Format.yaml:
        console.print(
            yaml.safe_dump(PublicKeys(pubkeys.public_keys).model_dump(mode="json", by_alias=True, exclude_unset=True))
        )


@pubkey_app.command(name="delete", rich_help_panel="Commands")
def delete_pubkey(ctx: Context, id: int):
    """Delete registered public key."""

    pubkey_api = ctx.obj.api.pubkey_api
    try:
        pubkey = pubkey_api.delete_pubkey(id)

        console.print(f"Deleted key {pubkey.id} with fingerprint {pubkey.sha256_fingerprint}")
    except NotFoundException:
        err_console.print("[bold red]No public key with this ID. :x:")
